/*
** Module   : P2JField.java
** Abstract : Helper class for dynamic temporary tables.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SVL 20130624 Created initial version.
*/

package com.goldencode.p2j.persist;

import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.xml.*;
import com.goldencode.util.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.atomic.*;

/**
 * Helper class for dynamic temporary tables.
 */
public class DynamicTablesHelper
{
   /** Base package for generated DMOs. */
   public final static String DMO_BASE_PACKAGE = "com.goldencode.p2j.persist.dynamic";

   /** Category name for DMO implementation classes, used in conversion rules. */
   private final static String DMO_IMPL_CATEGORY = "impl";

   /** Category name for DMO interfaces, used in conversion rules. */
   private final static String DMO_IFACE_CATEGORY = "iface";

   /** Suffix for DMO implementation classes. */
   private final static String DMO_IMPL_SUFFIX = "Impl";

   /** Separator used in generation of unique names for dynamic tables. */
   private final static String DYNAMI_DMO_SEPARATOR = "___";

   /** Singleton instance. */
   private static volatile DynamicTablesHelper instance;

   /** Counter used for generation of unique names for dynamic tables. */
   private final AtomicLong tableCounter = new AtomicLong(0);

   /** Convenience name converter instance. */
   private final NameConverter nc = new NameConverter();

   /**
    * Get singleton instance of the helper.
    *
    * @return  singleton instance of the helper.
    */
   public static DynamicTablesHelper getInstance()
   {
      if (instance == null)
      {
         synchronized (DynamicTablesHelper.class)
         {
            if (instance == null)
            {
               instance = new DynamicTablesHelper();
               PatternEngine.setDebugLevel(ConversionDriver.MSG_NONE);
            }
         }
      }

      return instance;
   }

   /**
    * Generates DMO interface, implementation class and mapping document. TODO
    *
    * @param builder
    *        Temporary table builder containing definitions of the target table.
    *
    * @return TODO
    */
   public boolean generateTable(TempTableBuilder builder)
   {
      long tableId = getUniqueTableId();

      // build table name
      StringBuilder sb = new StringBuilder(builder.getTableName());
      sb.append(DYNAMI_DMO_SEPARATOR);
      sb.append(tableId);
      String tableName = sb.toString();

      // build interface name
      sb = new StringBuilder(nc.convert(builder.getTableName(),
                                        MatchPhraseConstants.TYPE_CLASS));
      sb.append(DYNAMI_DMO_SEPARATOR);
      sb.append(tableId);
      String ifaceName = sb.toString();

      // build implementation class name
      sb = new StringBuilder(ifaceName);
      sb.append(DMO_IMPL_SUFFIX);
      String implClassName = sb.toString();

      DataModelAst root = generateDMOAst(builder, ifaceName, tableName);

      try
      {
         // generate .hbm.xml
         PatternEngine engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/hibernate");
         XmlAst xmlAst = (XmlAst) engine.getStoredObject(ifaceName);

         // generate interface and implementation Java ASTs
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/java_dmo");
         Aast impl = (Aast) engine.getStoredObject(DMO_IMPL_CATEGORY, ifaceName);
         Aast iface = (Aast) engine.getStoredObject(DMO_IFACE_CATEGORY, ifaceName);

         // generate Java code from Java ASTs
         engine = new PatternEngine();
         engine.addTargetAst(impl);
         engine.addTargetAst(iface);
         engine.run("convert/brew");

         // save output files
         String outputPath = "/home/svl/temp/";

         ByteArrayOutputStream implCodeStream = (ByteArrayOutputStream) engine.getStoredObject(ifaceName);
         ByteArrayOutputStream ifaceCodeStream = (ByteArrayOutputStream) engine.getStoredObject(implClassName);

         FileOutputStream output = new FileOutputStream(outputPath + ifaceName + ".java");
         output.write(ifaceCodeStream.toByteArray());
         output.close();

         output = new FileOutputStream(outputPath + implClassName + ".java");
         output.write(implCodeStream.toByteArray());
         output.close();

         Document dom = XmlAst.astToDom(xmlAst);
         XmlHelper.write(dom, new File(outputPath + ifaceName + ".hbm.xml"));
      }
      catch (ConfigurationException e)
      {
         e.printStackTrace();
      }
      catch (FileNotFoundException e)
      {
         e.printStackTrace();
      }
      catch (ParserConfigurationException e)
      {
         e.printStackTrace();
      }
      catch (IOException e)
      {
         e.printStackTrace();
      }

      return true;
   }

   /**
    * Generate and return a new unique id.
    *
    * @return  a new unique id.
    */
   private long getUniqueTableId()
   {
      return tableCounter.getAndIncrement();
   }

   /**
    * Generate AST that represents the target table and has the same structure as .p2o files from
    * the conversion stage.
    *
    * @param  builder
    *         Temporary table builder containing definitions of the target table.
    * @param  ifaceName
    *         The name of the target DMO interface.
    * @param  tableName
    *         The name of the target SQL table.
    *
    * @return the root node of the generated AST.
    */
   private DataModelAst generateDMOAst(TempTableBuilder builder,
                                       String ifaceName,
                                       String tableName)
   {
      DataModelAst root = new DataModelAst(new CommonToken(DataModelTokenTypes.DATA_MODEL,
                                                           DatabaseManager.TEMP_TABLE_SCHEMA));
      root.putAnnotation("package", DMO_BASE_PACKAGE);
      root.putAnnotation("permanent", true); //TODO

      DataModelAst cls = new DataModelAst(new CommonToken(DataModelTokenTypes.CLASS, ifaceName));
      cls.putAnnotation("table", tableName);
      cls.putAnnotation("historical", tableName);
      root.addChild(cls);

      DataModelAst primary = new DataModelAst(new CommonToken(DataModelTokenTypes.PRIMARY,
                                                              DatabaseManager.PRIMARY_KEY));
      primary.putAnnotation("column", DatabaseManager.PRIMARY_KEY);
      primary.putAnnotation("datatype", "Long");
      cls.addChild(primary);

      for (P2JField field : builder.getFields())
      {
         String fieldName = nc.convert(field.getName(), MatchPhraseConstants.TYPE_METHOD);

         DataModelAst property = new DataModelAst(new CommonToken(DataModelTokenTypes.PROPERTY,
                                                                  fieldName));
         property.putAnnotation("column", field.getName());
         property.putAnnotation("historical", field.getName());
         property.putAnnotation("datatype", field.getType());
         property.putAnnotation("label", field.getLabel());
         cls.addChild(property);
      }

      assignIds(root);

      return root;
   }

   /**
    * Traverse an AST and assign some unique ids to all of its nodes.
    *
    * @param root
    *        The root node of the AST.
    */
   private void assignIds(Aast root)
   {
      long id = 1;
      Iterator iter = root.iterator();
      while (iter.hasNext())
      {
         Aast ast = (Aast) iter.next();
         ast.setId(id++);
      }
   }
}
