/*
** Module   : TempTableBuilder.java
** Abstract : Implementation of the dynamic features for constructing temp-tables  
*              (attributes, methods and constructors).
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 OM  20130129 Created initial version.
** 002 VMN 20130217 Added support for handle-based attributes:
**                  ERROR (KW_ERROR)
** 003 CA  20130221 Added ADM-DATA, UNIQUE-ID, HandleChain and Deletable support.
** 004 CA  20130311 Added missing ADD-NEW-FIELD and ADD-NEW-INDEX API versions, using wrapper 
**                  types. Note that the conversion will use only the APIs with the wrapper type
**                  parameters; the other APIs (using non-wrapper parameters) are removed.
** 005 CS  20130322 Made the class non-abstract and added missing method stubs.
** 006 SVL 20130624 Some methods were implemented.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;
import java.util.*;

/**
 * Implementation of buffer of a dynamic TEMP-TABLE and the associated 
 * attributes and methods.
 * This {@link TempTable} interface is needed by {@link CommonHandle}
 * in order to store the handle to a the table.
 * 
 * @author om
 */
public class TempTableBuilder
extends HandleChain
implements TempTable,
           Deletable
{
   /**
    * Table name as it is specified in 4GL code, doesn't correspond Hibernate table or DMO class
    * name.
    */
   private String tableName;

   /** Table fields. */
   private final List<P2JField> fields = new ArrayList<P2JField>();

   /** Table indexes. */
   private final Map<String, P2JIndex> indexes = new LinkedHashMap<String, P2JIndex>();

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the new temp table, 
    *          if any, or empty string to create the object in the current 
    *          default widget pool.
    */
   public static void create(handle tth, String widgetPool)
   {
      tth.assign(new TempTableBuilder());
      // TODO add to widget pool
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    * 
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    * @param   widgetPool
    *          The name of the widget pool that contains the new temp table, 
    *          if any, or empty string to create the object in the current 
    *          default widget pool.
    */
   public static void create(handle tth, character widgetPool)
   {
      if (widgetPool.isUnknown())
      {
         TempTableBuilder.create(tth, "");
      }
      else
      {
         TempTableBuilder.create(tth, widgetPool.getValue());
      }
   }

   /**
    * Dynamically creates a temporary table at runtime and assign it to a 
    * handle. The table is empty and will be populated with fileds, indexes 
    * and relations subsequent statements.
    * This method is the P2J equivalent of <code>CREATE TEMP-TABLE</code> 
    * statement of Progress 4GL.
    *
    * @param   tth
    *          The handle to assign the newly created temporary table to.
    */
   public static void create(handle tth)
   {
      TempTableBuilder.create(tth, "");
   }

   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handler assigned to a table buffer to create the copy.
    *
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical createTableLike(handle hbuf)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    *
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical createTableLike(String tableName)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character name, character type)
   {
      return addNewField(name, type, null, null, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character name, character type, integer extent)
   {
      return addNewField(name, type, extent, null, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character name, character type, integer extent, character format)
   {
      return addNewField(name, type, extent, format, null, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial)
   {
      return addNewField(name, type, extent, format, initial, null, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label)
   {
      return addNewField(name, type, extent, format, initial, label, null);
   }

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * <p>
    * When a parameter is set to null, it means it was not specified by the business logic.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    * @param   columnLabel
    *          The label of the column associated with the defined field
    *          
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label,
                              character    columnLabel)
   {
      P2JField field = new P2JField(name.getValue(),
                                    type.getValue(),
                                    extent != null ? extent.getValue() : 0,
                                    format != null ? format.getValue() : null,
                                    initial,
                                    label != null ? label.getValue() : null,
                                    columnLabel != null ? columnLabel.getValue() : null);
      fields.add(field);
      return new logical(true);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name)
   {
      return addNewIndex(name, null, null, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name, logical unique)
   {
      return addNewIndex(name, unique, null, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name, logical unique, logical primary)
   {
      return addNewIndex(name, unique, primary, null);
   }

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    * <p>
    * If a parameter is null, it means it was not set.
    * 
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * @param   word
    *          TRUE if this is a word index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical addNewIndex(character name, logical unique, logical primary, logical word)
   {
      String indexName = name.getValue();
      P2JIndex index = new P2JIndex(tableName,
                                    indexName,
                                    unique != null && unique.getValue(),
                                    primary != null && primary.getValue(),
                                    word != null && word.getValue());
      indexes.put(indexName, index);
      return new logical(true);
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(String name, String source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(character name, character source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addFieldLike(character name, handle source)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field, String ascending)
   {
      return addFieldToIndex(new character(indexName), 
                             new character(field), new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, String field, character ascending)
   {
      return addFieldToIndex(new character(indexName), new character(field), ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, String field, String ascending)
   {
      return addFieldToIndex(indexName, new character(field), new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, String field, character ascending)
   {
      return addFieldToIndex(indexName, new character(field), ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, character field, String ascending)
   {
      return addFieldToIndex(new character(indexName), field, ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(String indexName, character field, character ascending)
   {
      return addFieldToIndex(new character(indexName), field, ascending);
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, character field, String ascending)
   {
      return addFieldToIndex(indexName, field, new character(ascending));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   public logical addFieldToIndex(character indexName, character field, character ascending)
   {
      // TODO implement logic for addFieldToIndex
      // TODO implement error handling for ascending not asc or desc
      // Error message: "Third argument to ADD-INDEX-FIELD must be DESC or ASC in quotes. (9062)"

      String ascVal = ascending.getValue();
      boolean asc = (ascVal == null);

      if (!asc)
      {
         ascVal = ascVal.toLowerCase();
         asc = ascVal.startsWith("asc");

         if (!asc)
         {
            if (!ascVal.startsWith("desc"))
            {
               ErrorManager.displayError("TODO Should be ASC or DESC");
               return new logical(false);
            }
         }
      }

      P2JIndex index = indexes.get(indexName.getValue());
      index.addComponent(field.getValue(), !asc, false, false, false);
      return new logical(true);
   }

   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(String indexName, String field)
   {
      return addFieldToIndex(new character(indexName), 
                             new character(field), new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(character indexName, String field)
   {
      return addFieldToIndex(indexName, new character(field), new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(String indexName, character field)
   {
      return addFieldToIndex(new character(indexName), field, new character("asc"));
   }
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   public logical addFieldToIndex(character indexName, character field)
   {
      return addFieldToIndex(indexName, field, new character("asc"));
   }

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addIndexLike(String name, String sourceName, String dbTable)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  TBA: Doc are not explicit, most likely true on success
    */
   public logical addIndexLike(character name,
                               character sourceName,
                               character dbTable)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Returns <code>true</code> if this temp-table is in prepared state (after a successful call 
    * to {@link #tempTablePrepare}). 
    * This method implements the read-only PREPARED attribute from 4GL.
    * 
    * @return  true if operation is successful 
    */
   public logical prepared()
   {
      // TODO implement actua logic
      return new logical(true);
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode. 
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful 
    */
   public logical copyTempTable(handle    other,
                                logical   append,
                                logical   replace,
                                logical   looseCopy,
                                character prefix)
   {
      // TODO implement actua logic
      return new logical(true);
   }

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode. 
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful 
    */
   public logical copyTempTable(handle  other,
                                boolean append,
                                boolean replace,
                                boolean looseCopy,
                                String  prefix)
   {
      // TODO implement actua logic
      return new logical(true);
   }
   /**
    * Sets the UNDO attribute of 4GL temp-table.
    *
    * @param   undo
    *          The new value for the attribute.
    */
   public void setCanUndo(logical undo)
   {
      // TODO implement actua logic
   }

   /** 
    * Obtain the current UNDO attribute of 4GL temp-table.
    * 
    * @return  true if the the temp-table can undo.
    */
   public logical canUndo()
   {
      // TODO implement actua logic
      return new logical(true);
   }

   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param   errStr
    *          The new value for the ERROR-STRING attribute.
    */
   public void changeErrorString(character errStr)
   {
      // TODO implement actua logic
   }

   /** 
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    * 
    * @return  the current ERROR-STRING of the temp-table
    */
   public character errorString()
   {
      // TODO implement actua logic
      return new character();
   }

   /**
    * Obtain the id for this temp-table. It is guaranteed to be unique.
    *  
    * @return  The unique id assigned to this object.
    */
   public integer uniqueId()
   {
      // TODO implement actua logic
      return new integer();
   }

   /**
    * Checks  whether the widget is dynamic or static.
    * 
    * @return  Always returns true for dynamically created temp-tables. 
    */
   public logical dynamic()
   {
      // TODO implement actua logic
      return new logical(true);
   }
   
   /**
    * Checks whether this temp-table contains any records.
    * 
    * @return  true if this temp-table has records.
    */
   public logical hasRecords()
   {
      // TODO implement actua logic
      return new logical(true);
   }

   /**
    * Returns the temp-table to it's clear state, removing all defined fields and indexes.
    * 
    * TODO: At the moment of writing this it's evident that the data from the table should also be
    * removed or method should fail (return false) if the table contains any data.
    * 
    * @return  true if operation is successful
    */
   public logical clear()
   {
      // TODO implement actua logic
      return new logical(true);
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical tempTablePrepare(String name)
   {
      tableName = name;
      boolean res = DynamicTablesHelper.getInstance().generateTable(this);
      return new logical(res);
   }

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *          
    * @return  TBA: Doc are not explicit, most likely true on success.
    */
   public logical tempTablePrepare(character name)
   {
      // TODO Auto-generated method stub
      return new logical();
   }
   
   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   buffh
    *          A handle to a buffer handle from which to copy the field definitions.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle buffh)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   buffh
    *          A handle to a buffer handle from which to copy the field definitions.
    * @param   except
    *          A list of fields names to be excluded in from the copy process.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(handle buffh, String... except)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A list of fields names to be excluded in from the copy process.
    * 
    * @return  <code>true</code> for success.
    */
   public logical addFieldsFrom(String tableName, String... except)
   {
      // TODO Auto-generated method stub
      return new logical();
   }

   /**
    * Obtain the handle to the buffer that was created by default for this 
    * temporary table. Every dynamic temp-table is created with at 
    * least one buffer. This buffer’s object handle is returned by this 
    * method. It cannot be called until the table has been prepared using 
    * <code>tempTablePrepare()</code> method.
    * This method is the P2J equivalent of <code>DEFAULT-BUFFER-HANDLE</code>
    * attribute.
    * 
    * @return  The default buffer handle for the temporary table.
    */
   public handle defaultBufferHandle()
   {
      // TODO Auto-generated method stub
      return new handle();
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public logical isError()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      // TODO:  implement
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      // TODO:  implement
   }

   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NAME attribute only in read-only mode.
    * 
    * @return   Always <code>false</code>, as NAME attribute is writable.
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return false;
   }

   /**
    * Gets the the unique ID number associated to this object by the
    * underlying system. Not the same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   public integer getUniqueID()
   {
      // TODO Auto-generated method stub
      return new integer();
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return true;
   }

   /**
    * Worker to be implemented by each resource.  Called by {@link #delete()}.
    */
   protected void resourceDelete()
   {
      // TODO Auto-generated method stub
      
   }
   
   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */   public character getADMData()
   {
      // TODO Auto-generated method stub
      return null;
   }

    /**
     * Set the value of the ADM-DATA attribute.
     * 
     * @param    value
     *           The new value.
     */
   public void setADMData(String value)
   {
      // TODO Auto-generated method stub
      
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   public void setADMData(character value)
   {
      // TODO Auto-generated method stub
      
   }

   /**
    * Get the table name as it is specified in 4GL code, doesn't correspond Hibernate table or
    * DMO class name.
    *
    * @return  the table name as it is specified in 4GL code.
    */
   public String getTableName()
   {
      return tableName;
   }

   /**
    * Get the list of table fields.
    *
    * @return  the list of table fields.
    */
   public List<P2JField> getFields()
   {
      return fields;
   }

   /**
    * Get table indexes.
    *
    * @return  the map of table indexes keyed by their names.
    */
   public Map<String, P2JIndex> getIndexes()
   {
      return indexes;
   }
}
